# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for creating Datasets by enumerating a template with a spreadsheet.

The templating code has specific expectations for how the reaction pbtxt and
spreadsheet are defined, namely that placeholder values in the pbtxt begin and
end with a "$" (dollar sign) and that these match a unique column header in the
spreadsheet file.
"""

import os
import re
from collections.abc import Mapping
from typing import BinaryIO, Optional, Union

import pandas as pd
from google.protobuf import text_format  # pytype: disable=import-error

import ord_schema
from ord_schema import validations
from ord_schema.proto import dataset_pb2, reaction_pb2

# pylint: disable=too-many-branches


def read_spreadsheet(file_name_or_buffer: Union[str, BinaryIO], suffix: Optional[str] = None) -> pd.DataFrame:
    """Reads a {csv, xls, xlsx} spreadsheet file.

    Args:
        file_name_or_buffer: Filename or buffer. Note that a buffer is only
            allowed if suffix is not None.
        suffix: Filename suffix, used to determine the data encoding.

    Returns:
        DataFrame containing the reaction spreadsheet data.
    """
    if suffix is None:
        _, suffix = os.path.splitext(file_name_or_buffer)
    if suffix in [".xls", ".xlsx"]:
        return pd.read_excel(file_name_or_buffer)
    return pd.read_csv(file_name_or_buffer)


def _is_null(value: Union[float, str]) -> bool:
    """Returns whether a value is null."""
    return pd.isnull(value) or (isinstance(value, str) and (value == "nan" or not value.strip()))


def _fill_template(string: str, substitutions: Mapping[str, ord_schema.ScalarType]) -> reaction_pb2.Reaction:
    """Performs substring substitutions according to a dictionary.

    If any pattern has a null replacement value (i.e. this is an empty cell in
    the spreadsheet), apply the following edits in the given order:

        1. If after templating, any identifier does not have a defined value,
           remove the identifier.
        2. If a compound doesn't have any identifiers, remove that compound.
        3. If a compound amount is NaN, remove that compound.
        4. If an input has no components, remove that input.

    These edits are logged for easier error checking.

    Note that these edits are focused on *inputs* and are not intended to be
    exhaustive (e.g. by covering all Compound messages anywhere in Reaction).

    Args:
        string: A string whose contents should be modified.
        substitutions: A dictionary where each (key, value) pair defines
            a substring to replace and what its replacement should be.

    Returns:
        Reaction message with substitutions filled in.

    Raises:
        ValueError: If the substituted reaction template cannot be parsed.
    """
    check_null = False
    for pattern, value in substitutions.items():
        if pd.isnull(value):
            check_null = True
        string = string.replace(pattern, repr(value).strip("'"))
    try:
        reaction = text_format.Parse(string, reaction_pb2.Reaction())
    except text_format.ParseError as error:
        raise ValueError(f"Failed to parse reaction pbtxt after templating: {error}") from error
    if check_null:
        for message in reaction.inputs.values():
            for component in message.components:
                for identifier in list(component.identifiers):
                    if _is_null(identifier.value):
                        component.identifiers.remove(identifier)
            for component in list(message.components):
                if not component.identifiers:
                    message.components.remove(component)
                    continue
                kind = component.amount.WhichOneof("kind")
                if kind is not None and _is_null(getattr(component.amount, kind).value):
                    message.components.remove(component)
        for key in list(reaction.inputs.keys()):
            if not reaction.inputs[key].components:
                del reaction.inputs[key]
    return reaction


def generate_dataset(template_string: str, df: pd.DataFrame, validate: bool = True) -> dataset_pb2.Dataset:
    """Generates a Dataset by enumerating a template reaction.

    Args:
        template_string: The contents of a Reaction pbtxt where placeholder
            values to be replaced are defined between dollar signs. For example,
            a SMILES identifier value could be "$product_smiles$". PLaceholders
            may only use letters, numbers, and underscores.
        df: Pandas Dataframe where each row corresponds to one reaction and
            column names match placeholders in the template_string.
        validate: Optional Boolean controlling whether Reaction messages should
            be validated as they are defined. Defaults to True.

    Returns:
        A Dataset message.

    Raises:
        ValueError: If there is no match for a placeholder string in df.
        ValueError: If validate is True and there are validation errors when
            validating an enumerated Reaction message.

    """
    placeholders = set(re.findall(r"\$\w+\$", template_string))
    for placeholder in placeholders:
        if placeholder not in df.columns:
            # Allow "$my_placeholder$" to match "my_placeholder" in df.
            if placeholder[1:-1] not in df.columns:
                raise ValueError(f"Placeholder {placeholder} not found as a" " column in dataset spreadsheet")
            df.rename(columns={placeholder[1:-1]: placeholder}, inplace=True)

    reactions = []
    for _, substitutions in df[list(placeholders)].iterrows():
        reaction = _fill_template(template_string, substitutions)
        # remove the reaction id field if reaction exits to avoid id conflicts
        if reaction.reaction_id:
            reaction.ClearField("reaction_id")

        if validate:
            output = validations.validate_message(reaction, raise_on_error=False)
            if output.errors:
                raise ValueError(f"Enumerated Reaction is not valid: {output.errors}")
        reactions.append(reaction)

    return dataset_pb2.Dataset(reactions=reactions)



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Macros for programmatic message creation."""
from typing import Optional, Union

from ord_schema import units
from ord_schema.proto import reaction_pb2

UNITS_RESOLVER = units.UnitResolver()
CONCENTRATION_RESOLVER = units.UnitResolver(units.CONCENTRATION_UNIT_SYNONYMS, forbidden_units={})


def simple_solution(
    solvent_smiles: str,
    solute_smiles: Optional[str] = None,
    volume: Union[None, str, reaction_pb2.Volume] = None,
    concentration: Union[None, str, reaction_pb2.Concentration] = None,
    saturated: bool = False,
) -> list[reaction_pb2.Compound]:
    """Creates a solution with at most one solvent and one solute.

    Args:
        solvent_smiles: SMILES of the solvent.
        solute_smiles: SMILES of the solute. If not specified, a pure solvent
            is generated.
        volume: Volume of solution (total volume including the solute).
        concentration: Concentration of solution. If both concentration and
            volume are specified, a solute quantity is computed. The schema
            does not currently have a way to represent an unknown volume of
            solution with a known concentration.
        saturated: Whether the solution is saturated. `Saturated` and
            `concentrated` cannot both be specified.
    Returns: A list of solvent/solute Compounds.
    """
    if saturated:
        if concentration is not None:
            raise ValueError("Cannot specify both `saturated=True` and a concentration.")
        if solute_smiles is None:
            raise ValueError("Must specify a solute if `saturated=True`")
    if isinstance(volume, str):
        volume = UNITS_RESOLVER.resolve(volume)
    if isinstance(concentration, str):
        concentration = CONCENTRATION_RESOLVER.resolve(concentration)

    if volume is not None and concentration is not None:
        solute_amount = units.compute_solute_quantity(volume, concentration)
    elif saturated:
        solute_amount = reaction_pb2.Amount(unmeasured=reaction_pb2.UnmeasuredAmount(type="SATURATED"))
    else:
        # This case covers 1. pure solvents 2. solution with unknown volume.
        solute_amount = None

    output_compounds = []
    solvent_pb = reaction_pb2.Compound()
    solvent_pb.identifiers.add(value=solvent_smiles, type="SMILES")
    if volume is not None:
        solvent_pb.amount.volume.MergeFrom(volume)
        if solute_smiles is not None:
            solvent_pb.amount.volume_includes_solutes = True
    output_compounds.append(solvent_pb)

    if solute_smiles:
        solute_pb = reaction_pb2.Compound()
        solute_pb.identifiers.add(value=solute_smiles, type="SMILES")
        if solute_amount is not None:
            solute_pb.amount.MergeFrom(solute_amount)
        output_compounds.append(solute_pb)
    return output_compounds


def brine(volume=None):
    return simple_solution(solute_smiles="[Na+].[Cl-]", solvent_smiles="O", volume=volume, saturated=True)



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Macros for quickly creating workup steps.

Example usage:

from ord_schema.macros import workups
from ord_schema.macros import solutions

The reaction was quenched with 300 ml of 0.5M NaHCO3 solution and the phases were separated.
The aqueous layer was back-extracted twice with 100 ml of EtOAc. The organic layers were
combined and washed with 200 ml of saturated NaCl solution, dried over MgSO4 and filtered.
Upon concentration to an oil, the crude product was purified by normal phase column
chromatography

reaction.workups.MergeFrom([
    workups.add_solution(solutions.simple_solution(
        solvent_smiles='O', solute_smiles='[Na+][HCO3-]', volume='300mL', concentration='0.5M')),
    workups.separate_phases(keep_phase='organic'),
    workups.add_solution(solutions.simple_solution(
        solvent_smiles='CC(=O)OCC', volume='100mL'), type='EXTRACTION'),
    workups.separate_phases(keep_phase='organic'),
    workups.add_solution(solutions.simple_solution(
        solvent_smiles='CC(=O)OCC', volume='100mL'), type='EXTRACTION'),
    workups.separate_phases(keep_phase='organic'),
    workups.add_solution(solutions.brine('200mL'), type='WASH'),
    workups.drying_agent('[Mg+2].[O-]S([O-])(=O)=O'),
    workups.filter(keep_phase='filtrate'),
    workups.rotovap(),
    reaction_pb2.ReactionWorkup(type='OTHER_CHROMATOGRAPHY'),
])
"""
from collections.abc import Iterable

from ord_schema import units
from ord_schema.proto import reaction_pb2

UNITS_RESOLVER = units.UnitResolver()
CONCENTRATION_RESOLVER = units.UnitResolver(units.CONCENTRATION_UNIT_SYNONYMS)


def add_solution(
    solution: Iterable[reaction_pb2.Compound], workup_type: str = "ADDITION"
) -> reaction_pb2.ReactionWorkup:
    """Create a workup representing addition of a solution.

    type is commonly one of 'ADDITION', 'EXTRACTION', or 'WASH'; see
    ReactionWorkup.WorkupType enum for full list of possible values.
    """
    workup = reaction_pb2.ReactionWorkup(type=workup_type)
    workup.input.components.MergeFrom(solution)
    for component in workup.input.components:
        component.reaction_role = reaction_pb2.ReactionRole.WORKUP
    return workup


def separate_phases(keep_phase: str) -> reaction_pb2.ReactionWorkup:
    """Create a workup representing a phase separation.

    keep_phase is commonly either 'aqueous' or 'organic'.
    """
    return reaction_pb2.ReactionWorkup(type="EXTRACTION", keep_phase=keep_phase)


def drying_agent(agent_smiles: str) -> reaction_pb2.ReactionWorkup:
    """Create a workup representing addition of a drying agent."""
    workup = reaction_pb2.ReactionWorkup(type="DRY_WITH_MATERIAL")
    component = workup.input.components.add()
    component.identifiers.add(value=agent_smiles, type="SMILES")
    component.reaction_role = reaction_pb2.ReactionRole.WORKUP
    return workup


def filtration(keep_phase: str) -> reaction_pb2.ReactionWorkup:
    """Create a workup representing a filtration step.

    keep_phase should be one of 'filtrate' or 'solid'.
    """
    return reaction_pb2.ReactionWorkup(type="FILTRATION", keep_phase=keep_phase)


def rotovap() -> reaction_pb2.ReactionWorkup:
    """Create a workup representing a rotary evaporation step."""
    return reaction_pb2.ReactionWorkup(type="CONCENTRATION")



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ord_schema.templating."""

import pandas as pd
import pytest
from google.protobuf import text_format
from rdkit import Chem

from ord_schema import templating
from ord_schema.proto import dataset_pb2, reaction_pb2


@pytest.fixture
def valid_reaction() -> reaction_pb2.Reaction:
    message = reaction_pb2.Reaction()
    dummy_input = message.inputs["in"]
    outcome = message.outcomes.add()
    component = dummy_input.components.add()
    component.identifiers.add(type="SMILES", value="CCO")
    component.is_limiting = True
    component.amount.mass.value = 1
    component.amount.mass.units = reaction_pb2.Mass.GRAM
    outcome.conversion.value = 75
    outcome.conversion.precision = 99
    message.provenance.record_created.time.value = "2023-07-01"
    message.provenance.record_created.person.name = "test"
    message.provenance.record_created.person.email = "test@example.com"
    yield message


def test_valid_templating(valid_reaction):
    template_string = text_format.MessageToString(valid_reaction)
    template_string = template_string.replace('value: "CCO"', 'value: "$smiles$"')
    template_string = template_string.replace("value: 75", "value: $conversion$")
    df = pd.DataFrame.from_dict({"$smiles$": ["CCO", "CCCO", "CCCCO"], "$conversion$": [75, 50, 30]})
    dataset = templating.generate_dataset(template_string, df)
    expected_reactions = []
    for smiles, conversion in zip(["CCO", "CCCO", "CCCCO"], [75, 50, 30]):
        reaction = reaction_pb2.Reaction()
        reaction.CopyFrom(valid_reaction)
        reaction.inputs["in"].components[0].identifiers[0].value = smiles
        reaction.outcomes[0].conversion.value = conversion
        expected_reactions.append(reaction)
    expected_dataset = dataset_pb2.Dataset(reactions=expected_reactions)
    assert dataset == expected_dataset

    # Test without "$" in column names
    df = pd.DataFrame.from_dict({"smiles": ["CCO", "CCCO", "CCCCO"], "conversion": [75, 50, 30]})
    dataset = templating.generate_dataset(template_string, df)
    assert dataset == expected_dataset


def test_valid_templating_escapes(valid_reaction):
    smiles = ["CCO", "CCCO", "CCCCO"]
    mols = [Chem.MolFromSmiles(this_smiles) for this_smiles in smiles]
    molblocks = [Chem.MolToMolBlock(mol) for mol in mols]
    valid_reaction.inputs["in"].components[0].identifiers.add(type="MOLBLOCK", value="$molblock$")
    template_string = text_format.MessageToString(valid_reaction)
    df = pd.DataFrame.from_dict({"molblock": molblocks})
    dataset = templating.generate_dataset(template_string, df)
    expected_reactions = []
    for molblock in molblocks:
        reaction = reaction_pb2.Reaction()
        reaction.CopyFrom(valid_reaction)
        reaction.inputs["in"].components[0].identifiers[1].value = molblock
        expected_reactions.append(reaction)
    expected_dataset = dataset_pb2.Dataset(reactions=expected_reactions)
    assert dataset == expected_dataset


@pytest.mark.parametrize("suffix", [".csv", ".xlsx"])
def test_read_spreadsheet(suffix, tmp_path):
    df = pd.DataFrame.from_dict({"smiles": ["CCO", "CCCO", "CCCCO"], "conversion": [75, 50, 30]})
    filename = (tmp_path / f"test{suffix}").as_posix()
    if suffix == ".csv":
        df.to_csv(filename, index=False)
    else:
        df.to_excel(filename, index=False)
    pd.testing.assert_frame_equal(templating.read_spreadsheet(filename), df)


def test_invalid_templating(valid_reaction):
    template_string = text_format.MessageToString(valid_reaction)
    template_string = template_string.replace('value: "CCO"', 'value: "$my_smiles$"')
    template_string = template_string.replace("precision: 99", "precision: $precision$")
    df = pd.DataFrame.from_dict({"$my_smiles$": ["CCO", "CCCO", "CCCCO"], "$precision$": [75, 50, -5]})
    expected_reactions = []
    for smiles, precision in zip(["CCO", "CCCO", "CCCCO"], [75, 50, -5]):
        reaction = reaction_pb2.Reaction()
        reaction.CopyFrom(valid_reaction)
        reaction.inputs["in"].components[0].identifiers[0].value = smiles
        reaction.outcomes[0].conversion.precision = precision
        expected_reactions.append(reaction)
    expected_dataset = dataset_pb2.Dataset(reactions=expected_reactions)
    with pytest.raises(ValueError, match="Enumerated Reaction is not valid"):
        templating.generate_dataset(template_string, df)
    dataset = templating.generate_dataset(template_string, df, validate=False)
    assert dataset == expected_dataset


def test_bad_placeholders(valid_reaction):
    template_string = text_format.MessageToString(valid_reaction)
    template_string = template_string.replace('value: "CCO"', 'value: "$my_smiles$"')
    template_string = template_string.replace("value: 75", "value: $conversion$")
    df = pd.DataFrame.from_dict({"$my_smiles$": ["CCO", "CCCO", "CCCCO"]})
    with pytest.raises(ValueError, match=r"\$conversion\$ not found"):
        templating.generate_dataset(template_string, df)


def test_missing_values(tmp_path):
    # pylint: disable=too-many-locals
    # Build a template reaction.
    reaction = reaction_pb2.Reaction()
    reaction.provenance.record_created.time.value = "2023-07-01"
    reaction.provenance.record_created.person.name = "test"
    reaction.provenance.record_created.person.email = "test@example.com"
    input1 = reaction.inputs["one"]
    input1_component1 = input1.components.add()
    input1_component1.identifiers.add(value="CCO", type="SMILES")
    input1_component1.amount.mass.value = 1.2
    input1_component1.amount.mass.units = reaction_pb2.Mass.GRAM
    input1_component2 = input1.components.add()
    input1_component2.is_limiting = True
    input1_component2.identifiers.add(value="c1ccccc1", type="SMILES")
    input1_component2.amount.volume.value = 3.4
    input1_component2.amount.volume.units = reaction_pb2.Volume.LITER
    input2 = reaction.inputs["two"]
    input2_component1 = input2.components.add()
    input2_component1.identifiers.add(value="COO", type="SMILES")
    input2_component1.amount.mass.value = 5.6
    input2_component1.amount.mass.units = reaction_pb2.Mass.GRAM
    outcome = reaction.outcomes.add()
    outcome.conversion.value = 75
    template_string = text_format.MessageToString(reaction)
    template_string = template_string.replace('value: "CCO"', 'value: "$smiles$"')
    template_string = template_string.replace("value: 5.6", "value: $mass$")
    # Build a spreadsheet and test for proper edits.
    filename = (tmp_path / "missing.csv").as_posix()
    with open(filename, "w") as f:
        f.write("smiles,mass\n")
        f.write("CN,\n")  # Missing mass.
        f.write(",1.5\n")  # Missing SMILES.
    df = pd.read_csv(filename)
    dataset = templating.generate_dataset(template_string, df)
    expected_dataset = dataset_pb2.Dataset()
    reaction1 = expected_dataset.reactions.add()
    reaction1.CopyFrom(reaction)
    reaction1.inputs["one"].components[0].identifiers[0].value = "CN"
    del reaction1.inputs["two"]  # No components after amount removal.
    reaction2 = expected_dataset.reactions.add()
    reaction2.CopyFrom(reaction)
    del reaction2.inputs["one"].components[0]  # No indentifiers.
    reaction2.inputs["two"].components[0].amount.mass.value = 1.5
    assert dataset == expected_dataset



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Name/string resolution to structured messages or identifiers."""
import email.message
import re
import urllib.error
import urllib.parse
import urllib.request

from rdkit import Chem

from ord_schema import message_helpers
from ord_schema.logging import get_logger
from ord_schema.proto import reaction_pb2

logger = get_logger(__name__)

_COMPOUND_STRUCTURAL_IDENTIFIERS = [
    reaction_pb2.CompoundIdentifier.SMILES,
    reaction_pb2.CompoundIdentifier.INCHI,
    reaction_pb2.CompoundIdentifier.MOLBLOCK,
    reaction_pb2.CompoundIdentifier.CXSMILES,
    reaction_pb2.CompoundIdentifier.XYZ,
]

_USERNAME = "github-actions"
_EMAIL = "github-actions@github.com"


def canonicalize_smiles(smiles: str) -> str:
    """Canonicalizes a SMILES string.

    Args:
        smiles: SMILES string.

    Returns:
        Canonicalized SMILES string.

    Raises:
        ValueError: If the SMILES cannot be parsed by RDKit.
    """
    mol = Chem.MolFromSmiles(smiles)
    if not mol:
        raise ValueError(f"Could not parse SMILES: {smiles}")
    return Chem.MolToSmiles(mol)


def name_resolve(value_type: str, value: str) -> tuple[str, str]:
    """Resolves compound identifiers to SMILES via multiple APIs."""
    for resolver, resolver_func in _NAME_RESOLVERS.items():
        try:
            smiles = resolver_func(value_type, value)
            if smiles is not None:
                return smiles, resolver
        except urllib.error.HTTPError as error:
            logger.info(f"{resolver} could not resolve {value_type} {value}: {error}")
    raise ValueError(f"Could not resolve {value_type} {value} to SMILES")


def resolve_names(message: reaction_pb2.Reaction) -> bool:
    """Attempts to resolve compound NAME identifiers to SMILES.

    When a NAME identifier is resolved, a SMILES identifier is added to the list
    of identifiers for that compound. Note that this function moves on to the
    next Compound after the first successful name resolution.

    Args:
        message: Reaction proto.

    Returns:
        Boolean whether `message` was modified.
    """
    modified = False
    compounds = message_helpers.find_submessages(message, reaction_pb2.Compound)
    for compound in compounds:
        if any(identifier.type in _COMPOUND_STRUCTURAL_IDENTIFIERS for identifier in compound.identifiers):
            continue  # Compound already has a structural identifier.
        for identifier in compound.identifiers:
            if identifier.type == identifier.NAME:
                try:
                    smiles, resolver = name_resolve("name", identifier.value)
                    new_identifier = compound.identifiers.add()
                    new_identifier.type = new_identifier.SMILES
                    new_identifier.value = smiles
                    new_identifier.details = f"NAME resolved by the {resolver}"
                    modified = True
                    break
                except ValueError:
                    pass
    return modified


def _pubchem_resolve(value_type: str, value: str) -> str:
    """Resolves compound identifiers to SMILES via the PubChem REST API."""
    with urllib.request.urlopen(
        f"https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/{value_type}/"
        f"{urllib.parse.quote(value)}/property/IsomericSMILES/txt"
    ) as response:
        return response.read().decode().strip()


def _cactus_resolve(value_type: str, value: str) -> str:
    """Resolves compound identifiers to SMILES via the CACTUS API."""
    del value_type  # Unused.
    with urllib.request.urlopen(
        "https://cactus.nci.nih.gov/chemical/structure/" f"{urllib.parse.quote(value)}/smiles"
    ) as response:
        return response.read().decode().strip()


def _emolecules_resolve(value_type: str, value: str) -> str:
    """Resolves compound identifiers to SMILES via the eMolecules API."""
    del value_type  # Unused.
    with urllib.request.urlopen("https://www.emolecules.com/" f"lookup?q={urllib.parse.quote(value)}") as response:
        response_text = response.read().decode().strip()
    if response_text == "__END__":
        raise urllib.error.HTTPError("", 404, "eMolecules lookup unsuccessful", email.message.Message(), None)
    return response_text.split("\t")[0]


def resolve_input(input_string: str) -> reaction_pb2.ReactionInput:
    """Resolve a text-based description of an input in one of the following
    formats:
        (1) [AMOUNT] of [NAME]
        (2) [AMOUNT] of [CONCENTRATION] [SOLUTE] in [SOLVENT]

    Args:
        input_string: String describing the input.

    Returns:
        ReactionInput message.

    Raises:
        ValueError: if the string cannot be parsed properly.
    """
    reaction_input = reaction_pb2.ReactionInput()
    if " of " not in input_string:
        raise ValueError("String does not match template!")
    amount_string, description = input_string.split(" of ")
    if " in " not in description:
        component_name = description
        component = reaction_input.components.add()
        component.CopyFrom(message_helpers.build_compound(name=component_name.strip(), amount=amount_string))
        resolve_names(reaction_input)
        return reaction_input
    pattern = re.compile(r"(\d+.?\d*)\s?(\w+)\s(.+)\sin\s(.+)")
    match = pattern.fullmatch(description.strip())
    if not match:
        raise ValueError("String did not match template!")
    conc_value, conc_units, solute_name, solvent_name = match.groups()
    assert solute_name is not None  # Type hint.
    assert solvent_name is not None  # Type hint.
    solute = reaction_input.components.add()
    solvent = reaction_input.components.add()
    solute.CopyFrom(message_helpers.build_compound(name=solute_name.strip()))
    solvent.CopyFrom(message_helpers.build_compound(name=solvent_name.strip(), amount=amount_string))
    if solvent.amount.WhichOneof("kind") != "volume":
        raise ValueError("Total amount of solution must be a volume!")
    solvent.amount.volume_includes_solutes = True
    message_helpers.set_solute_moles(solute, [solvent], f"{conc_value} {conc_units}")
    resolve_names(reaction_input)
    return reaction_input


# Standard name resolvers.
_NAME_RESOLVERS = {
    "PubChem API": _pubchem_resolve,
    "NCI/CADD Chemical Identifier Resolver": _cactus_resolve,
    "eMolecules Lookup Service": _emolecules_resolve,
}



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ord_schema.updates."""

import pytest

from ord_schema import updates
from ord_schema.proto import dataset_pb2, reaction_pb2


class TestUpdateReaction:
    def test_with_updates_simple(self):
        message = reaction_pb2.Reaction()
        updates.update_reaction(message)
        assert message != reaction_pb2.Reaction()
        assert len(message.provenance.record_modified) == 1

    def test_with_no_updates(self):
        message = reaction_pb2.Reaction()
        message.provenance.record_created.time.value = "2020-05-08"
        message.reaction_id = "ord-c0bbd41f095a44a78b6221135961d809"
        copied = reaction_pb2.Reaction()
        copied.CopyFrom(message)
        updates.update_reaction(copied)
        assert copied == message

    def test_add_reaction_id(self):
        message = reaction_pb2.Reaction()
        updates.update_reaction(message)
        assert message.reaction_id
        assert len(message.provenance.record_modified) == 1

    def test_keep_existing_reaction_id(self):
        message = reaction_pb2.Reaction()
        message.reaction_id = "ord-c0bbd41f095a44a78b6221135961d809"
        message.provenance.record_created.time.value = "2020-01-01"
        updates.update_reaction(message)
        assert message.reaction_id == "ord-c0bbd41f095a44a78b6221135961d809"
        assert len(message.provenance.record_modified) == 0


class TestUpdateDataset:
    @pytest.fixture
    def dataset(self) -> dataset_pb2.Dataset:
        dataset = dataset_pb2.Dataset()
        reaction = dataset.reactions.add()
        # Minimal reaction.
        dummy_input = reaction.inputs["dummy_input"]
        reaction.outcomes.add()
        dummy_component = dummy_input.components.add()
        dummy_component.identifiers.add(type="CUSTOM")
        dummy_component.identifiers[0].details = "custom_identifier"
        dummy_component.identifiers[0].value = "custom_value"
        dummy_component.amount.mass.value = 1
        dummy_component.amount.mass.units = reaction_pb2.Mass.GRAM
        # Placeholders for referenced reactions.
        reaction2 = dataset.reactions.add()
        reaction3 = dataset.reactions.add()
        reaction2.CopyFrom(dataset.reactions[0])
        reaction3.CopyFrom(dataset.reactions[0])
        yield dataset

    def test_crossferences(self, dataset):
        dummy_input = dataset.reactions[0].inputs["dummy_input"]
        dummy_input.crude_components.add(reaction_id="crude", has_derived_amount=True)
        dataset.reactions[1].reaction_id = "crude"
        dummy_input.components[0].preparations.add(type="SYNTHESIZED", reaction_id="synthesized")
        dataset.reactions[2].reaction_id = "synthesized"
        # Check updated values.
        updates.update_dataset(dataset)
        assert dummy_input.crude_components[0].reaction_id == dataset.reactions[1].reaction_id
        assert dummy_input.crude_components[0].reaction_id != "crude"
        assert dummy_input.components[0].preparations[0].reaction_id == dataset.reactions[2].reaction_id
        assert dummy_input.components[0].preparations[0].reaction_id != "synthesized"

    def test_crossferences_are_proper_ids(self, dataset):
        dummy_input = dataset.reactions[0].inputs["dummy_input"]
        dummy_input.crude_components.add(reaction_id="ord-c0bbd41f095a44a78b6221135961d809", has_derived_amount=True)
        dataset.reactions[1].reaction_id = "ord-c0bbd41f095a44a78b6221135961d809"
        dummy_input.components[0].preparations.add(
            type="SYNTHESIZED", reaction_id="ord-d0bbd41f095a44a78b6221135961d809"
        )
        dataset.reactions[2].reaction_id = "ord-d0bbd41f095a44a78b6221135961d809"
        # Check updated values.
        updates.update_dataset(dataset)
        assert dummy_input.crude_components[0].reaction_id == dataset.reactions[1].reaction_id
        assert dummy_input.crude_components[0].reaction_id == "ord-c0bbd41f095a44a78b6221135961d809"
        assert dummy_input.components[0].preparations[0].reaction_id == dataset.reactions[2].reaction_id
        assert dummy_input.components[0].preparations[0].reaction_id == "ord-d0bbd41f095a44a78b6221135961d809"



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RDKit PostgreSQL cartridge functionality.

Notes:
  * These tables live in a separate "rdkit" schema to avoid name conflicts between tables and extension types.
  * The RDKit-specific columns are populated by ord_schema.orm.database.add_rdkit; this allows the ORM to function
    normally even if if the RDKit PostgreSQL cartridge is not installed (the `smiles` column will be populated and
    the other columns will be empty).
  * Objects with this type are added to the ORM in from_proto() using the `rdkit` field.
"""
from __future__ import annotations

import os
from distutils.util import strtobool  # pylint: disable=deprecated-module
from enum import Enum

from sqlalchemy import Column, Index, Integer, Text, cast, func
from sqlalchemy.sql.expression import ColumnElement
from sqlalchemy.types import UserDefinedType

from ord_schema.orm import Base


def rdkit_cartridge() -> bool:
    """Returns whether to use RDKit PostgreSQL cartridge functionality."""
    return bool(strtobool(os.environ.get("ORD_POSTGRES_RDKIT", "1")))


class RDKitMol(UserDefinedType):
    """https://github.com/rdkit/rdkit/blob/master/Code/PgSQL/rdkit/rdkit.sql.in#L4."""

    cache_ok = True

    @property
    def python_type(self):
        raise NotImplementedError

    def get_col_spec(self, **kwargs):
        """Returns the column type."""
        del kwargs  # Unused.
        return "mol" if rdkit_cartridge() else "bytea"


class RDKitReaction(UserDefinedType):
    """https://github.com/rdkit/rdkit/blob/master/Code/PgSQL/rdkit/rdkit.sql.in#L129."""

    cache_ok = True

    @property
    def python_type(self):
        raise NotImplementedError

    def get_col_spec(self, **kwargs):
        """Returns the column type."""
        del kwargs  # Unused.
        return "reaction" if rdkit_cartridge() else "bytea"


class RDKitBfp(UserDefinedType):
    """https://github.com/rdkit/rdkit/blob/master/Code/PgSQL/rdkit/rdkit.sql.in#L81."""

    cache_ok = True

    @property
    def python_type(self):
        raise NotImplementedError

    def get_col_spec(self, **kwargs):
        """Returns the column type."""
        del kwargs  # Unused.
        return "bfp" if rdkit_cartridge() else "bytea"


class RDKitSfp(UserDefinedType):
    """https://github.com/rdkit/rdkit/blob/master/Code/PgSQL/rdkit/rdkit.sql.in#L105."""

    cache_ok = True

    @property
    def python_type(self):
        raise NotImplementedError

    def get_col_spec(self, **kwargs):
        """Returns the column type."""
        del kwargs  # Unused.
        return "sfp" if rdkit_cartridge() else "bytea"


class CString(UserDefinedType):
    """PostgreSQL cstring."""

    cache_ok = True

    @property
    def python_type(self):
        raise NotImplementedError

    def get_col_spec(self, **kwargs):
        """Returns the column type."""
        del kwargs  # Unused.
        return "cstring"


class FingerprintType(Enum):
    """RDKit PostgreSQL fingerprint types."""

    # NOTE(skearnes): Add Column and Index entries for each member to RDKitMol below.
    MORGAN_BFP = func.morganbv_fp
    MORGAN_SFP = func.morgan_fp

    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)


class RDKitMols(Base):
    """Table for storing compound structures and associated RDKit cartridge data."""

    __tablename__ = "mols"
    id = Column(Integer, primary_key=True)
    smiles = Column(Text, index=True, unique=True)
    mol = Column(RDKitMol)
    morgan_bfp = Column(RDKitBfp)
    morgan_sfp = Column(RDKitSfp)

    __table_args__ = (
        Index("mol_index", "mol", postgresql_using="gist"),
        Index("morgan_bfp_index", "morgan_bfp", postgresql_using="gist"),
        Index("morgan_sfp_index", "morgan_sfp", postgresql_using="gist"),
        {"schema": "rdkit"},
    )

    @classmethod
    def tanimoto(cls, other: str, fp_type: FingerprintType = FingerprintType.MORGAN_BFP) -> ColumnElement[float]:
        return func.tanimoto_sml(getattr(cls, fp_type.name.lower()), fp_type(cast(other, RDKitMol)))

    @classmethod
    def contains_substructure(cls, pattern: str) -> ColumnElement[bool]:
        return func.substruct(cls.mol, cast(pattern, RDKitMol))

    @classmethod
    def matches_smarts(cls, pattern: str) -> ColumnElement[bool]:
        return func.substruct(cls.mol, func.qmol_from_smarts(cast(pattern, CString)))


class RDKitReactions(Base):
    """Table for storing reaction objects and associated RDKit cartridge data."""

    __tablename__ = "reactions"
    id = Column(Integer, primary_key=True)
    reaction_smiles = Column(Text, index=True, unique=True)
    reaction = Column(RDKitReaction)

    __table_args__ = (
        Index("reaction_index", "reaction", postgresql_using="gist"),
        {"schema": "rdkit"},
    )

    @classmethod
    def matches_smarts(cls, pattern: str) -> ColumnElement[bool]:
        return func.substruct(cls.reaction, func.reaction_from_smarts(cast(pattern, CString)))



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pytest fixtures."""
import os
import re
from typing import Iterator

import pytest
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from testing.postgresql import Postgresql

from ord_schema.message_helpers import load_message
from ord_schema.orm.database import add_dataset, prepare_database
from ord_schema.proto import dataset_pb2


@pytest.fixture(name="test_engine")
def test_engine_fixture() -> Iterator[Engine]:
    with Postgresql() as postgres:
        # See https://docs.sqlalchemy.org/en/20/dialects/postgresql.html#module-sqlalchemy.dialects.postgresql.psycopg.
        url = re.sub("postgresql://", "postgresql+psycopg://", postgres.url())
        engine = create_engine(url, future=True)
        yield engine


@pytest.fixture(name="test_session")
def test_session_fixture(test_engine) -> Iterator[Session]:
    datasets = [
        load_message(
            os.path.join(os.path.dirname(__file__), "testdata", "ord-nielsen-example.pbtxt"), dataset_pb2.Dataset
        )
    ]
    rdkit_cartridge = prepare_database(test_engine)
    with Session(test_engine) as session:
        for dataset in datasets:
            with session.begin():
                add_dataset(dataset, session, rdkit_cartridge=rdkit_cartridge)
    with Session(test_engine) as session:
        yield session



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Builds a Dataset from a set of Reaction protos.

TODO(kearnes): Add support for automatic sharding?

Usage:
    build_dataset.py --input=<str> --output=<str> --name=<str> --description=<str> [--no-validate]

Options:
    --input=<str>           Input pattern for Reaction protos
    --output=<str>          Output Dataset filename (*.pbtxt)
    --name=<str>            Name for this dataset
    --description=<str>     Description for this dataset
    --no-validate           If set, do run validations on reactions
"""
import glob

import docopt

from ord_schema import message_helpers, validations
from ord_schema.logging import get_logger
from ord_schema.proto import dataset_pb2, reaction_pb2

logger = get_logger(__name__)


def main(kwargs):
    filenames = glob.glob(kwargs["--input"], recursive=True)
    logger.info("Found %d Reaction protos", len(filenames))
    reactions = []
    for filename in filenames:
        reactions.append(message_helpers.load_message(filename, reaction_pb2.Reaction))
    dataset = dataset_pb2.Dataset(name=kwargs["--name"], description=kwargs["--description"], reactions=reactions)
    if not kwargs["--no-validate"]:
        validations.validate_datasets({"_COMBINED": dataset})
    message_helpers.write_message(dataset, kwargs["--output"])


if __name__ == "__main__":
    main(docopt.docopt(__doc__))



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Adds datasets to the ORM database.

Usage:
    add_datasets.py --pattern=<str> [options]
    add_datasets.py -h | --help

Options:
    --pattern=<str>         Pattern for dataset filenames
    --overwrite             Update changed datasets
    --url=<str>             Postgres connection string
    --database=<str>        Database [default: orm]
    --username=<str>        Database username [default: postgres]
    --password=<str>        Database password
    --host=<str>            Database host [default: localhost]
    --port=<int>            Database port [default: 5432]
    --n_jobs=<int>          Number of parallel workers [default: 1]
"""
import os
from concurrent.futures import ProcessPoolExecutor, as_completed
from glob import glob
from hashlib import md5

from docopt import docopt
from rdkit import RDLogger
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from tqdm import tqdm

from ord_schema.logging import get_logger
from ord_schema.message_helpers import load_message
from ord_schema.orm import database
from ord_schema.proto import dataset_pb2

engine: Engine = None
logger = get_logger(__name__)


def initializer():
    """Initializer for child processes."""
    # See https://docs.sqlalchemy.org/en/20/core/pooling.html#using-connection-pools-with-multiprocessing-or-os-fork.
    engine.dispose(close=False)


def add_dataset(filename: str, overwrite: bool) -> str:
    """Adds a single dataset to the database.

    Args:
        filename: Dataset filename.
        overwrite: If True, update the dataset if the MD5 hash has changed.

    Returns:
        Dataset ID.

    Raises:
        ValueError: If the dataset already exists in the database and `overwrite` is not set.
    """
    logger.debug(f"Loading {filename}")
    dataset = load_message(filename, dataset_pb2.Dataset)
    with Session(engine) as session:
        with session.begin():
            dataset_md5 = database.get_dataset_md5(dataset.dataset_id, session)
        if dataset_md5 is not None:
            this_md5 = md5(dataset.SerializeToString(deterministic=True)).hexdigest()
            if this_md5 != dataset_md5:
                if not overwrite:
                    raise ValueError(f"`overwrite` is required when a dataset already exists: {dataset.dataset_id}")
                logger.debug(f"existing dataset {dataset.dataset_id} changed; updating")
                with session.begin():
                    database.delete_dataset(dataset.dataset_id, session)
            else:
                logger.debug(f"existing dataset {dataset.dataset_id} unchanged; skipping")
                return dataset.dataset_id
        with session.begin():
            database.add_dataset(dataset, session, rdkit_cartridge=False)  # Do this separately in add_rdkit().
    return dataset.dataset_id


def add_rdkit(dataset_id: str) -> None:
    """Updates RDKit tables."""
    with Session(engine) as session:
        with session.begin():
            database.update_rdkit_tables(dataset_id, session)
        with session.begin():
            database.update_rdkit_ids(dataset_id, session)


def main(**kwargs):
    RDLogger.DisableLog("rdApp.*")
    if kwargs.get("--url"):
        url = kwargs["--url"]
    else:
        url = database.get_connection_string(
            database=kwargs["--database"],
            username=kwargs["--username"],
            password=kwargs["--password"] or os.environ["PGPASSWORD"],
            host=kwargs["--host"],
            port=int(kwargs["--port"]),
        )
    global engine  # pylint: disable=global-statement
    engine = create_engine(url)
    filenames = sorted(glob(kwargs["--pattern"]))
    with ProcessPoolExecutor(initializer=initializer, max_workers=int(kwargs["--n_jobs"])) as executor:
        logger.info("Adding datasets")
        futures = {}
        for filename in filenames:
            future = executor.submit(add_dataset, filename=filename, overwrite=kwargs["--overwrite"])
            futures[future] = filename
        dataset_ids = []
        failures = []
        for future in tqdm(as_completed(futures), total=len(futures)):
            try:
                dataset_ids.append(future.result())
            except Exception as error:  # pylint: disable=broad-exception-caught
                filename = futures[future]
                failures.append(filename)
                logger.error(f"Adding dataset {filename} failed: {error}")
        logger.info("Adding RDKit functionality")
        for dataset_id in tqdm(dataset_ids):
            try:
                add_rdkit(dataset_id)  # NOTE(skearnes): Do this serially to avoid deadlocks.
            except Exception as error:  # pylint: disable=broad-exception-caught
                failures.append(dataset_id)
                logger.error(f"Adding RDKit functionality for {dataset_id} failed: {error}")
        if failures:
            raise RuntimeError(failures)


if __name__ == "__main__":
    main(**docopt(__doc__))



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Automated updates for Reaction messages."""

import datetime
import re
import uuid

from ord_schema.proto import dataset_pb2, reaction_pb2

_COMPOUND_STRUCTURAL_IDENTIFIERS = [
    reaction_pb2.CompoundIdentifier.SMILES,
    reaction_pb2.CompoundIdentifier.INCHI,
    reaction_pb2.CompoundIdentifier.MOLBLOCK,
    reaction_pb2.CompoundIdentifier.CXSMILES,
    reaction_pb2.CompoundIdentifier.XYZ,
]

_USERNAME = "github-actions"
_EMAIL = "github-actions@github.com"


def update_reaction(reaction: reaction_pb2.Reaction) -> dict[str, str]:
    """Updates a Reaction message.

    Current updates:
      * Sets reaction_id if not already set.
      * Adds a record modification event to the provenance.

    Args:
        reaction: reaction_pb2.Reaction message.

    Returns:
        A dictionary mapping placeholder reaction_ids to newly-assigned
            reaction_ids.
    """
    modified = False
    id_substitutions = {}
    if not re.fullmatch("^ord-[0-9a-f]{32}$", reaction.reaction_id):
        # NOTE(kearnes): This does not check for the case where a Dataset is
        # edited and reaction_id values are changed inappropriately. This will
        # need to be either (1) caught in review or (2) found by a complex
        # check of the diff.
        reaction_id = f"ord-{uuid.uuid4().hex}"
        if reaction.reaction_id:
            id_substitutions[reaction.reaction_id] = reaction_id
        reaction.reaction_id = reaction_id
        modified = True
    for func in _UPDATES:
        # NOTE(kearnes): Order is important here; if you write
        # `modified or func(reaction)` and modified is True, the interpreter
        # will skip the evaluation of func(reaction).
        modified = func(reaction) or modified
    if modified:
        event = reaction.provenance.record_modified.add()
        event.time.value = datetime.datetime.utcnow().ctime()
        event.person.username = _USERNAME
        event.person.email = _EMAIL
        event.details = "Automatic updates from the submission pipeline."
    return id_substitutions


def update_dataset(dataset: dataset_pb2.Dataset):
    """Updates a Dataset message.

    Current updates:
      * All reaction-level updates in update_reaction.
      * reaction_id cross-references between Reactions in the dataset.

    Args:
        dataset: dataset_pb2.Dataset message.

    Raises:
        KeyError: if the dataset has not been validated and there exists a
            cross-referenced reaction_id in any Reaction that is not defined
            elsewhere in the Dataset.
    """
    if not re.fullmatch("^ord_dataset-[0-9a-f]{32}$", dataset.dataset_id):
        dataset.dataset_id = f"ord_dataset-{uuid.uuid4().hex}"
    # Reaction-level updates
    id_substitutions = {}
    for reaction in dataset.reactions:
        id_substitutions.update(update_reaction(reaction))
    # Dataset-level updates of cross-references
    for reaction in dataset.reactions:
        for reaction_input in reaction.inputs.values():
            for component in reaction_input.components:
                for preparation in component.preparations:
                    if preparation.reaction_id and preparation.reaction_id in id_substitutions:
                        preparation.reaction_id = id_substitutions[preparation.reaction_id]
            for crude_component in reaction_input.crude_components:
                if crude_component.reaction_id in id_substitutions:
                    crude_component.reaction_id = id_substitutions[crude_component.reaction_id]


# Standard updates.
_UPDATES = []



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for ord_schema.orm.scripts.add_datasets."""
import os

import docopt
import pytest

from ord_schema.orm.database import prepare_database
from ord_schema.orm.scripts import add_datasets


def test_main(test_engine):
    if not prepare_database(test_engine):
        pytest.skip("RDKit cartridge is required")
    argv = [
        "--url",
        test_engine.url,
        "--pattern",
        os.path.join(os.path.dirname(__file__), "..", "testdata", "ord-nielsen-example.pbtxt"),
    ]
    add_datasets.main(**docopt.docopt(add_datasets.__doc__, argv))



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for creating/managing the PostgreSQL database."""
import os
import time
from unittest.mock import patch

from sqlalchemy import cast, delete, func, select, text, update
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.engine import Engine
from sqlalchemy.exc import NotSupportedError, OperationalError
from sqlalchemy.orm import Session

from ord_schema.logging import get_logger
from ord_schema.orm.mappers import Base, Mappers, from_proto
from ord_schema.orm.rdkit_mappers import CString, FingerprintType, RDKitMols, RDKitReactions
from ord_schema.proto import dataset_pb2

logger = get_logger(__name__)


def get_connection_string(
    database: str, username: str, password: str, host: str = "localhost", port: int = 5432
) -> str:
    """Creates an SQLAlchemy connection string."""
    return f"postgresql+psycopg://{username}:{password}@{host}:{port}/{database}"


def prepare_database(engine: Engine) -> bool:
    """Prepares the database and creates the ORM table structure.

    Args:
        engine: SQLAlchemy Engine.

    Returns:
        Whether the RDKit PostgreSQL cartridge is installed.
    """
    with engine.begin() as connection:
        try:
            connection.execute(text("CREATE EXTENSION IF NOT EXISTS tsm_system_rows"))  # For random sampling.
        except OperationalError:
            logger.warning("tsm_system_rows cartridge is not installed; random sampling will be disabled")
    with engine.begin() as connection:
        connection.execute(text("CREATE SCHEMA IF NOT EXISTS ord"))
        connection.execute(text("CREATE SCHEMA IF NOT EXISTS rdkit"))
    try:
        with engine.begin() as connection:
            # NOTE(skearnes): The RDKit PostgreSQL extension works best in the public schema.
            connection.execute(text("CREATE EXTENSION IF NOT EXISTS rdkit"))
        rdkit_cartridge = True
    except (OperationalError, NotSupportedError):
        with engine.begin() as connection:
            logger.warning("RDKit PostgreSQL cartridge is not installed; structure search will be disabled")
            connection.execute(text("CREATE EXTENSION IF NOT EXISTS btree_gist"))
        rdkit_cartridge = False
    with patch.dict(os.environ, {"ORD_POSTGRES_RDKIT": "1" if rdkit_cartridge else "0"}):
        Base.metadata.create_all(engine)
    return rdkit_cartridge


def add_dataset(dataset: dataset_pb2.Dataset, session: Session, rdkit_cartridge: bool = True) -> None:
    """Adds a dataset to the database."""
    logger.debug(f"Adding dataset {dataset.dataset_id}")
    start = time.time()
    mapped_dataset = from_proto(dataset)
    logger.debug(f"from_proto() took {time.time() - start:g}s")
    session.add(mapped_dataset)
    if rdkit_cartridge:
        session.flush()
        update_rdkit_tables(dataset.dataset_id, session)
        session.flush()
        update_rdkit_ids(dataset.dataset_id, session)


def get_dataset_md5(dataset_id: str, session: Session) -> str | None:
    """Returns the MD5 hash of the current version of a dataset, if it exists in the database."""
    result = session.execute(select(Mappers.Dataset.md5).where(Mappers.Dataset.dataset_id == dataset_id))
    row = result.first()
    return row[0] if row else None


def get_dataset_size(dataset_id: str, session: Session) -> int:
    """Returns the number of reactions in a dataset."""
    result = session.execute(select(Mappers.Dataset.num_reactions).where(Mappers.Dataset.dataset_id == dataset_id))
    row = result.first()
    if row is None:
        raise ValueError(dataset_id)
    return row[0]


def delete_dataset(dataset_id: str, session: Session) -> None:
    """Deletes a dataset from the database."""
    logger.debug(f"Deleting dataset {dataset_id}")
    start = time.time()
    session.execute(delete(Mappers.Dataset).where(Mappers.Dataset.dataset_id == dataset_id))
    logger.debug(f"delete took {time.time() - start}s")


def update_rdkit_tables(dataset_id: str, session: Session) -> None:
    """Updates RDKit PostgreSQL cartridge data."""
    _update_rdkit_reactions(dataset_id, session)
    _update_rdkit_mols(dataset_id, session)


def _update_rdkit_reactions(dataset_id: str, session: Session) -> None:
    """Updates the RDKit reactions table."""
    logger.debug("Updating RDKit reactions")
    assert hasattr(RDKitReactions, "__table__")  # Type hint.
    table = RDKitReactions.__table__
    start = time.time()
    session.execute(
        insert(table)
        .from_select(
            ["reaction_smiles"],
            select(Mappers.Reaction.reaction_smiles)
            .join(Mappers.Dataset)
            .where(Mappers.Dataset.dataset_id == dataset_id, Mappers.Reaction.reaction_smiles.is_not(None))
            .distinct(),
        )
        .on_conflict_do_nothing(index_elements=["reaction_smiles"])
    )
    logger.debug(f"Updating reaction SMILES took {time.time() - start:g}s")
    start = time.time()
    session.execute(
        update(table)
        .where(table.c.reaction.is_(None))
        .values(reaction=func.reaction_from_smiles(cast(table.c.reaction_smiles, CString)))
    )
    logger.debug(f"reaction_from_smiles took {time.time() - start:g}s")


def _update_rdkit_mols(dataset_id: str, session: Session) -> None:
    """Updates the RDKit mols table."""
    logger.debug("Updating RDKit mols")
    assert hasattr(RDKitMols, "__table__")  # Type hint.
    table = RDKitMols.__table__
    start = time.time()
    # NOTE(skearnes): This join path does not include non-input compounds like workups, internal standards, etc.
    session.execute(
        insert(table)
        .from_select(
            ["smiles"],
            select(Mappers.Compound.smiles)
            .join(Mappers.ReactionInput)
            .join(Mappers.Reaction)
            .join(Mappers.Dataset)
            .where(
                Mappers.Dataset.dataset_id == dataset_id,
                Mappers.Compound.smiles.is_not(None),
                # See https://github.com/open-reaction-database/ord-schema/issues/672.
                Mappers.Compound.smiles.not_like("%[Ti+5]%"),
            )
            .distinct(),
        )
        .on_conflict_do_nothing(index_elements=["smiles"])
    )
    session.execute(
        insert(table)
        .from_select(
            ["smiles"],
            select(Mappers.ProductCompound.smiles)
            .join(Mappers.ReactionOutcome)
            .join(Mappers.Reaction)
            .join(Mappers.Dataset)
            .where(Mappers.Dataset.dataset_id == dataset_id, Mappers.ProductCompound.smiles.is_not(None))
            .distinct(),
        )
        .on_conflict_do_nothing(index_elements=["smiles"])
    )
    logger.debug(f"Updating SMILES took {time.time() - start:g}s")
    start = time.time()
    session.execute(
        update(table).where(table.c.mol.is_(None)).values(mol=func.mol_from_smiles(cast(table.c.smiles, CString)))
    )
    logger.debug(f"mol_from_smiles took {time.time() - start:g}s")
    logger.debug("Updating fingerprints")
    for fp_type in FingerprintType:
        start = time.time()
        column = fp_type.name.lower()
        session.execute(
            update(table)
            .where(getattr(table.c, column).is_(None), table.c.mol.is_not(None))
            .values(**{column: fp_type(table.c.mol)})
        )
        logger.debug(f"{fp_type} took {time.time() - start:g}s")


def update_rdkit_ids(dataset_id: str, session: Session) -> None:
    """Updates RDKit reaction and mol ID associations in the ORD tables."""
    logger.debug("Updating RDKit ID associations")
    start = time.time()
    # Update Reaction.
    query = session.execute(
        select(Mappers.Reaction.id, RDKitReactions.id)
        .join(RDKitReactions, Mappers.Reaction.reaction_smiles == RDKitReactions.reaction_smiles)
        .join(Mappers.Dataset)
        .where(Mappers.Dataset.dataset_id == dataset_id)
    )
    updates = []
    for ord_id, rdkit_id in query.fetchall():
        updates.append({"id": ord_id, "rdkit_reaction_id": rdkit_id})
    session.execute(update(Mappers.Reaction), updates)
    # Update Compound.
    query = session.execute(
        select(Mappers.Compound.id, RDKitMols.id)
        .join(RDKitMols, Mappers.Compound.smiles == RDKitMols.smiles)
        .join(Mappers.ReactionInput)
        .join(Mappers.Reaction)
        .join(Mappers.Dataset)
        .where(Mappers.Dataset.dataset_id == dataset_id)
    )
    updates = []
    for ord_id, rdkit_id in query.fetchall():
        updates.append({"id": ord_id, "rdkit_mol_id": rdkit_id})
    session.execute(update(Mappers.Compound), updates)
    # Update ProductCompound.
    query = session.execute(
        select(Mappers.ProductCompound.id, RDKitMols.id)
        .join(RDKitMols, Mappers.ProductCompound.smiles == RDKitMols.smiles)
        .join(Mappers.ReactionOutcome)
        .join(Mappers.Reaction)
        .join(Mappers.Dataset)
        .where(Mappers.Dataset.dataset_id == dataset_id)
    )
    updates = []
    for ord_id, rdkit_id in query.fetchall():
        updates.append({"id": ord_id, "rdkit_mol_id": rdkit_id})
    session.execute(update(Mappers.ProductCompound), updates)
    logger.debug(f"Updating RDKit IDs took {time.time() - start:g}s")



from ord_schema.proto import reaction_pb2 as _reaction_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Dataset(_message.Message):
    __slots__ = ["name", "description", "reactions", "reaction_ids", "dataset_id"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    REACTIONS_FIELD_NUMBER: _ClassVar[int]
    REACTION_IDS_FIELD_NUMBER: _ClassVar[int]
    DATASET_ID_FIELD_NUMBER: _ClassVar[int]
    name: str
    description: str
    reactions: _containers.RepeatedCompositeFieldContainer[_reaction_pb2.Reaction]
    reaction_ids: _containers.RepeatedScalarFieldContainer[str]
    dataset_id: str
    def __init__(self, name: _Optional[str] = ..., description: _Optional[str] = ..., reactions: _Optional[_Iterable[_Union[_reaction_pb2.Reaction, _Mapping]]] = ..., reaction_ids: _Optional[_Iterable[str]] = ..., dataset_id: _Optional[str] = ...) -> None: ...

class DatasetExample(_message.Message):
    __slots__ = ["dataset_id", "description", "url", "created"]
    DATASET_ID_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    CREATED_FIELD_NUMBER: _ClassVar[int]
    dataset_id: str
    description: str
    url: str
    created: _reaction_pb2.RecordEvent
    def __init__(self, dataset_id: _Optional[str] = ..., description: _Optional[str] = ..., url: _Optional[str] = ..., created: _Optional[_Union[_reaction_pb2.RecordEvent, _Mapping]] = ...) -> None: ...



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ord_schema.scripts.build_dataset."""
import os

import docopt
import pytest

from ord_schema import message_helpers, validations
from ord_schema.proto import dataset_pb2, reaction_pb2
from ord_schema.scripts import build_dataset


@pytest.fixture
def dirname(tmp_path) -> str:
    reaction1 = reaction_pb2.Reaction()
    dummy_input = reaction1.inputs["dummy_input"]
    dummy_component = dummy_input.components.add()
    dummy_component.identifiers.add(type="CUSTOM")
    dummy_component.identifiers[0].details = "custom_identifier"
    dummy_component.identifiers[0].value = "custom_value"
    dummy_component.is_limiting = True
    dummy_component.amount.mass.value = 1
    dummy_component.amount.mass.units = reaction_pb2.Mass.GRAM
    reaction1.outcomes.add().conversion.value = 75
    reaction1.provenance.record_created.time.value = "2023-07-01"
    reaction1.provenance.record_created.person.name = "test"
    reaction1.provenance.record_created.person.email = "test@example.com"
    dirname = tmp_path.as_posix()
    message_helpers.write_message(reaction1, os.path.join(dirname, "reaction-1.pbtxt"))
    # reaction2 is empty.
    reaction2 = reaction_pb2.Reaction()
    message_helpers.write_message(reaction2, os.path.join(dirname, "reaction-2.pbtxt"))
    yield dirname


def test_simple(dirname):
    input_pattern = os.path.join(dirname, "reaction-1.pbtxt")
    output_filename = os.path.join(dirname, "dataset.pbtxt")
    argv = [
        "--input",
        input_pattern,
        "--name",
        "test dataset",
        "--description",
        "this is a test dataset",
        "--output",
        output_filename,
    ]
    build_dataset.main(docopt.docopt(build_dataset.__doc__, argv))
    assert os.path.exists(output_filename)
    dataset = message_helpers.load_message(output_filename, dataset_pb2.Dataset)
    assert dataset.name == "test dataset"
    assert dataset.description == "this is a test dataset"
    assert len(dataset.reactions) == 1


def test_validation(dirname):
    input_pattern = os.path.join(dirname, "reaction-?.pbtxt")
    output_filename = os.path.join(dirname, "dataset.pbtxt")
    argv = [
        "--input",
        input_pattern,
        "--name",
        "test dataset",
        "--description",
        "this is a test dataset",
        "--output",
        output_filename,
    ]
    with pytest.raises(
        validations.ValidationError,
        match="Reactions should have at least 1 reaction input",
    ):
        build_dataset.main(docopt.docopt(build_dataset.__doc__, argv))
    # Make sure disabling validation works.
    argv = [
        "--input",
        input_pattern,
        "--name",
        "test dataset",
        "--description",
        "this is a test dataset",
        "--output",
        output_filename,
        "--no-validate",
    ]
    build_dataset.main(docopt.docopt(build_dataset.__doc__, argv))



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ord_schema.units."""

import pytest

from ord_schema import units
from ord_schema.proto import reaction_pb2


@pytest.fixture
def resolver() -> units.UnitResolver:
    return units.UnitResolver()


@pytest.mark.parametrize(
    "string,expected",
    (
        ("15.0 ML", reaction_pb2.Volume(value=15.0, units=reaction_pb2.Volume.MILLILITER)),
        ("24 H", reaction_pb2.Time(value=24, units=reaction_pb2.Time.HOUR)),
        ("32.1g", reaction_pb2.Mass(value=32.1, units=reaction_pb2.Mass.GRAM)),
        ("   32.1      \t   g  ", reaction_pb2.Mass(value=32.1, units=reaction_pb2.Mass.GRAM)),
        ("10 min.", reaction_pb2.Time(value=10.0, units=reaction_pb2.Time.MINUTE)),
        ("-78°C", reaction_pb2.Temperature(value=-78.0, units=reaction_pb2.Temperature.CELSIUS)),
        ("10±5g", reaction_pb2.Mass(value=10, precision=5, units=reaction_pb2.Mass.GRAM)),
        (" 10 meter", reaction_pb2.Length(value=10, units=reaction_pb2.Length.METER)),
        ("1.2e-3g", reaction_pb2.Mass(value=0.0012, units=reaction_pb2.Mass.GRAM)),
        ("0.12 nL", reaction_pb2.Volume(value=0.12, units=reaction_pb2.Volume.NANOLITER)),
    ),
)
def test_resolve(resolver, string, expected):
    assert resolver.resolve(string) == expected


@pytest.mark.parametrize(
    "message,new_units,expected",
    (
        (
            reaction_pb2.Volume(value=15.0, units=reaction_pb2.Volume.MILLILITER),
            "L",
            reaction_pb2.Volume(value=0.015, units=reaction_pb2.Volume.LITER),
        ),
        (
            reaction_pb2.Time(value=24, units=reaction_pb2.Time.HOUR),
            "s",
            reaction_pb2.Time(value=86400, units=reaction_pb2.Time.SECOND),
        ),
        (
            reaction_pb2.Time(value=24, units=reaction_pb2.Time.HOUR),
            reaction_pb2.Time.SECOND,
            reaction_pb2.Time(value=86400, units=reaction_pb2.Time.SECOND),
        ),
        (
            reaction_pb2.Temperature(value=-78.0, units=reaction_pb2.Temperature.CELSIUS),
            "K",
            reaction_pb2.Temperature(value=195.15, units=reaction_pb2.Temperature.KELVIN),
        ),
        (
            reaction_pb2.Temperature(value=-78.0, units=reaction_pb2.Temperature.CELSIUS, precision=5.2),
            "K",
            reaction_pb2.Temperature(value=195.15, units=reaction_pb2.Temperature.KELVIN, precision=5.2),
        ),
        (
            reaction_pb2.Temperature(value=25, units=reaction_pb2.Temperature.CELSIUS, precision=5),
            "F",
            reaction_pb2.Temperature(value=77, units=reaction_pb2.Temperature.FAHRENHEIT, precision=9),
        ),
        (
            reaction_pb2.Mass(value=0.0012, units=reaction_pb2.Mass.GRAM),
            "mg",
            reaction_pb2.Mass(value=1.2, units=reaction_pb2.Mass.MILLIGRAM),
        ),
        (
            reaction_pb2.Wavelength(value=500, units=reaction_pb2.Wavelength.NANOMETER),
            "cm^-1",
            reaction_pb2.Wavelength(value=20000, units=reaction_pb2.Wavelength.WAVENUMBER),
        ),
        (
            reaction_pb2.Moles(value=0.0003, units=reaction_pb2.Moles.MOLE),
            "umol",
            reaction_pb2.Moles(value=300, units=reaction_pb2.Moles.MICROMOLE),
        ),
        (
            reaction_pb2.Pressure(value=1, units=reaction_pb2.Pressure.ATMOSPHERE, precision=0.1),
            "bar",
            reaction_pb2.Pressure(value=1.01325, units=reaction_pb2.Pressure.BAR, precision=0.101325),
        ),
        (
            reaction_pb2.FlowRate(value=2, units=reaction_pb2.FlowRate.MILLILITER_PER_SECOND),
            "ul/s",
            reaction_pb2.FlowRate(value=2000, units=reaction_pb2.FlowRate.MICROLITER_PER_SECOND),
        ),
        (
            reaction_pb2.Voltage(value=50, units=reaction_pb2.Voltage.VOLT, precision=2),
            "mv",
            reaction_pb2.Voltage(value=50000, units=reaction_pb2.Voltage.MILLIVOLT, precision=2000),
        ),
    ),
)
def test_convert(resolver, message, new_units, expected):
    assert resolver.convert(message, new_units) == expected


@pytest.mark.parametrize(
    "message,new_units,expected",
    (
        (reaction_pb2.Volume(value=15.0, units=reaction_pb2.Volume.MILLILITER), "smoot", "unrecognized units"),
        (reaction_pb2.Volume(value=15.0, units=reaction_pb2.Volume.MILLILITER), "gram", "different types"),
        (reaction_pb2.Temperature(value=5, units=reaction_pb2.Temperature.CELSIUS), "meter", "different types"),
    ),
)
def test_convert_should_fail(resolver, message, new_units, expected):
    with pytest.raises((KeyError, ValueError), match=expected):
        resolver.convert(message, new_units)


@pytest.mark.parametrize(
    "volume,concentration,expected",
    (("1L", "1 molar", "1 mol"), ("3mL", "0.1 molar", "300 micromoles"), ("100mL", "0.1 molar", "10 millimoles")),
)
def test_compute_solute_quantity(resolver, volume, concentration, expected):
    conc_resolver = units.UnitResolver(unit_synonyms=units.CONCENTRATION_UNIT_SYNONYMS)
    assert units.compute_solute_quantity(
        volume=resolver.resolve(volume),
        concentration=conc_resolver.resolve(concentration),
    ) == reaction_pb2.Amount(moles=resolver.resolve(expected))


@pytest.mark.parametrize(
    "string,expected",
    (("1-2 h", reaction_pb2.Time(value=1.5, precision=0.5, units=reaction_pb2.Time.HOUR)),),
)
def test_resolve_allow_range(resolver, string, expected):
    assert resolver.resolve(string, allow_range=True) == expected


@pytest.mark.parametrize(
    "string,expected",
    (
        ("1.21 GW", "unrecognized units"),
        ("15.0 ML 20.0 L", "string does not contain a value with units"),
        ("15.0. ML", "string does not contain a value with units"),
        ("5.2 m", "ambiguous"),
    ),
)
def test_resolve_should_fail(resolver, string, expected):
    with pytest.raises((KeyError, ValueError), match=expected):
        resolver.resolve(string)
